/* 
 * Copyright (C) 2001-2007 Jacek Sieka, arnetheduck on gmail point com
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifndef USER_H
#define USER_H

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Util.h"
#include "Pointer.h"
#include "CID.h"
#include "CriticalSection.h"

/** A user connected to one or more hubs. */
class User : public PointerBase, public Flags8
{
public:
	/** Each flag is set if it's true in at least one hub */
	enum UserFlags {
		ONLINE        = 0x1,
		DCPLUSPLUS    = 0x2,
		PASSIVE       = 0x4,
		TTH_GET       = 0x8,		//< User supports getting files by tth -> don't have path in queue...
		SAVE_NICK     = 0x10	//< Save cid->nick association
	};

	typedef Pointer<User> Ptr;
	typedef vector<Ptr> List;
	typedef List::iterator Iter;

	struct HashFunction {
#ifdef _MSC_VER
		static const size_t bucket_size = 4;
		static const size_t min_buckets = 8;
#endif
		size_t operator()(const Ptr& x) const { return ((size_t)(&(*x)))/sizeof(User); }
		bool operator()(const Ptr& a, const Ptr& b) const { return (&(*a)) < (&(*b)); }
	};

	User(const CID& aCID) : cid(aCID) { }

	virtual ~User() throw() { }

	const CID& getCID() const { return cid; }
	operator const CID&() const { return cid; }

	bool isOnline() const { return isSet(ONLINE); }

	GETSET(string, firstNick, FirstNick);
private:
	User(const User&);
	User& operator=(const User&);

	CID cid;
};

/** One of possibly many identities of a user, mainly for UI purposes */
class Identity : public Flags8 {
public:
	enum IdentityBits {
		OP        = 0x1, 
		REG       = 0x2, 
		AWAY      = 0x4,
		SERVER    = 0x8,
		FIREBALL  = 0x10, 
		RCTM_SND  = 0x20
	};

	Identity() { }
	Identity(const User::Ptr& ptr) : user(ptr), ctype(11) { }
	Identity(const Identity& rhs) : Flags8(rhs), user(rhs.user), info(rhs.info), ctype(rhs.ctype) { }
	Identity& operator=(const Identity& rhs) { Lock l(cs); *static_cast<Flags8*>(this) = rhs; user = rhs.user; info = rhs.info; ctype = rhs.ctype; return *this; }
		
#define GS(n, x) string get##n() const { return get(x); } void set##n(const string& v) { set(x, v); }
	GS(Nick, "NI")
	GS(Description, "DE")
	GS(Ip, "I4")
	GS(Email, "EM")
	GS(Connection, "CO")
	GS(ClientID, "CL")
	GS(Version, "VE")
	GS(Mode, "MD")
	GS(Hubs, "HT")
	GS(Slots, "SL")
	GS(Upload, "UL")

	void setBytesShared(const string& bs) { set("SS", bs); }
	int64_t getBytesShared() const { return Util::toInt64(get("SS")); }
	
	void setOp(bool op) { op ? setFlag(OP) : unsetFlag(OP); }
	string getTag() const;
	bool isOp() const { return isSet(OP); }
	bool isRegistered() const { return isSet(REG); }

	string get(const char* name) const;
	void set(const char* name, const string& val);

	void getParams(StringMap& map, const string& prefix, bool compatibility) const;
	User::Ptr& getUser() { return user; }
	GETSET(User::Ptr, user, User);
	GETSET(uint8_t, ctype, cType);
private:
	typedef map<short, string> InfMap;
	typedef InfMap::iterator InfIter;
	InfMap info;
	/** @todo there are probably more threading issues here ...*/
	mutable CriticalSection cs;
};

class Client;
class NmdcHub;

class OnlineUser {
public:
	typedef vector<OnlineUser*> List;
	typedef List::iterator Iter;

	OnlineUser(const User::Ptr& ptr, Client& client_);
	virtual ~OnlineUser() throw() { }

	operator User::Ptr&() { return getUser(); }
	operator const User::Ptr&() const { return getUser(); }
	
	User::Ptr& getUser() { return getIdentity().getUser(); }
	const User::Ptr& getUser() const { return getIdentity().getUser(); }
	Identity& getIdentity() { return identity; }
	Client& getClient() { return client; }
	const Client& getClient() const { return client; }

	GETSET(Identity, identity, Identity);
private:
	friend class NmdcHub;
	
	OnlineUser(const OnlineUser&);
	OnlineUser& operator=(const OnlineUser&);

	Client& client;
};

#endif // USER_H

/* Some parts Copyright (C) 2002-2007 Petr Kozelka, PPK at czdc dot org */
